@@ -1,5 +1,5 @@
 /*
- * $Id: pgmodule.c,v 1.59 2005/09/01 11:41:36 darcy Exp $
+ * $Id: pgmodule.c,v 1.73 2006/06/01 23:14:11 cito Exp $
  * PyGres, version 2.2 A Python interface for PostgreSQL database. Written by
  * D'Arcy J.M. Cain, (darcy@druid.net).  Based heavily on code written by
  * Pascal Andre, andre@chimay.via.ecp.fr. Copyright (c) 1995, Pascal Andre
@@ -27,16 +27,22 @@
  *
  */
 
-#include <Python.h>
+/* Note: This should be linked against the same C runtime lib as Python */
+
 #include "postgres.h"
 #include "libpq-fe.h"
 #include "libpq/libpq-fs.h"
 #include "catalog/pg_type.h"
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
 
-/* PyObject_Del does not exist in older versions of Python. */
+/* these will be defined in Python.h again: */
+#undef _POSIX_C_SOURCE
+#undef HAVE_STRERROR
+#undef snprintf
+#undef vsnprintf
+
+#include <Python.h>
+
+/* PyObject_Del does not exist in older versions of Python: */
 #if PY_VERSION_HEX < 0x01060000
 #define PyObject_Del(op) PyMem_DEL((op))
 #endif
@@ -45,7 +51,7 @@ static PyObject *Error, *Warning, *InterfaceError,
 	*DatabaseError, *InternalError, *OperationalError, *ProgrammingError,
 	*IntegrityError, *DataError, *NotSupportedError;
 
-static const char *PyPgVersion = "3.7";
+static const char *PyPgVersion = "3.8.1";
 
 /* taken from fileobject.c */
 #define BUF(v) PyString_AS_STRING((PyStringObject *)(v))
@@ -91,10 +97,6 @@ const char *__movename[5] =
 #define DEFAULT_VARS	1		/* enables default variables use */
 #endif   /* NO_DEF_VAR */
 
-#ifdef MS_WIN32
-#define NO_SNPRINTF 1
-#endif
-
 /* In 7.4 PQfreeNotify was deprecated and PQfreemem is used instead.  A
    macro exists in 7.4 for backwards compatibility. */
 #ifndef PQfreeNotify	/* must be earlier than 7.4 */
@@ -114,10 +116,10 @@ static PyObject *pg_default_tty;	/* default debug tty */
 static PyObject *pg_default_port;		/* default connection port */
 static PyObject *pg_default_user;		/* default username */
 static PyObject *pg_default_passwd;		/* default password */
-#endif   /* DEFAULT_VARS */
+#endif	/* DEFAULT_VARS */
 
 DL_EXPORT(void) init_pg(void);
-int		   *get_type_array(PGresult *result, int nfields);
+int *get_type_array(PGresult *result, int nfields);
 
 /* --------------------------------------------------------------------- */
 /* OBJECTS DECLARATION */
@@ -128,8 +130,8 @@ typedef struct
 {
 	PyObject_HEAD
 	int			valid;			/* validity flag */
-	PGconn	   *cnx;			/* PostGres connection handle */
-	PGresult   *last_result;	/* last result content */
+	PGconn		*cnx;			/* PostGres connection handle */
+	PGresult	*last_result;	/* last result content */
 }	pgobject;
 
 staticforward PyTypeObject PgType;
@@ -139,7 +141,7 @@ staticforward PyTypeObject PgType;
 static PyObject *
 pgobject_New(void)
 {
-	pgobject   *pgobj;
+	pgobject	*pgobj;
 
 	if ((pgobj = PyObject_NEW(pgobject, &PgType)) == NULL)
 		return NULL;
@@ -155,7 +157,7 @@ pgobject_New(void)
 typedef struct
 {
 	PyObject_HEAD
-	PGresult   *last_result;	/* last result content */
+	PGresult	*last_result;	/* last result content */
 	int			result_type;	/* type of previous result */
 	long		current_pos;	/* current position in last result */
 	long		num_rows;		/* number of (affected) rows */
@@ -171,8 +173,8 @@ typedef struct
 {
 	PyObject_HEAD
 	int			valid;			/* validity flag */
-	pgobject   *pgcnx;			/* parent connection object */
-	PGresult   *last_result;	/* last result content */
+	pgobject	*pgcnx;			/* parent connection object */
+	PGresult	*last_result;	/* last result content */
 	int			result_type;	/* result type (DDL/DML/DQL) */
 	long		arraysize;		/* array size for fetch method */
 	int			current_row;	/* current selected row */
@@ -191,7 +193,7 @@ staticforward PyTypeObject PgSourceType;
 typedef struct
 {
 	PyObject_HEAD
-	pgobject * pgcnx;			/* parent connection object */
+	pgobject	*pgcnx;			/* parent connection object */
 	Oid			lo_oid;			/* large object oid */
 	int			lo_fd;			/* large object fd */
 }	pglargeobject;
@@ -204,13 +206,107 @@ staticforward PyTypeObject PglargeType;
 /* --------------------------------------------------------------------- */
 /* INTERNAL FUNCTIONS */
 
+
+/* prints result (mostly useful for debugging) */
+/* Note: This is a simplified version of the Postgres function PQprint().
+ * PQprint() is not used because handing over a stream from Python to
+ * Postgres can be problematic if they use different libs for streams.
+ * Also, PQprint() is considered obsolete and may be removed sometime.
+ */
+static void
+print_result(FILE *fout, const PGresult *res)
+{
+	int n = PQnfields(res);
+	if (n > 0)
+	{
+		int i, j;
+		int *fieldMax = NULL;
+		char **fields = NULL;
+		const char **fieldNames;
+		int m = PQntuples(res);
+		if (!(fieldNames = (const char **) calloc(n, sizeof(char *))))
+		{
+			fprintf(stderr, "out of memory\n"); exit(1);
+		}
+		if (!(fieldMax = (int *) calloc(n, sizeof(int))))
+		{
+			fprintf(stderr, "out of memory\n"); exit(1);
+		}
+		for (j = 0; j < n; j++)
+		{
+			const char *s = PQfname(res, j);
+			fieldNames[j] = s;
+			fieldMax[j] = s ? strlen(s) : 0;
+		}
+		if (!(fields = (char **) calloc(n * (m + 1), sizeof(char *))))
+		{
+			fprintf(stderr, "out of memory\n"); exit(1);
+		}
+		for (i = 0; i < m; i++)
+		{
+			for (j = 0; j < n; j++)
+			{
+				const char *val;
+				int len;
+				len = PQgetlength(res, i, j);
+				val = PQgetvalue(res, i, j);
+				if (len >= 1 && val && *val)
+				{
+					if (len > fieldMax[j])
+						fieldMax[j] = len;
+					if (!(fields[i * n + j] = (char *) malloc(len + 1)))
+					{
+						fprintf(stderr, "out of memory\n"); exit(1);
+					}
+					strcpy(fields[i * n + j], val);
+				}
+			}
+		}
+		for (j = 0; j < n; j++)
+		{
+			const char *s = PQfname(res, j);
+			int len = strlen(s);
+			if (len > fieldMax[j])
+				fieldMax[j] = len;
+			fprintf(fout, "%-*s", fieldMax[j], s);
+			if (j + 1 < n)
+				fputc('|', fout);
+		}
+		fputc('\n', fout);
+		for (j = 0; j < n; j++)
+		{
+			for (i = fieldMax[j]; i--; fputc('-', fout));
+			if (j + 1 < n)
+				fputc('+', fout);
+		}
+		fputc('\n', fout);
+		for (i = 0; i < m; i++)
+		{
+			for (j = 0; j < n; j++)
+			{
+				char *s = fields[i * n + j];
+				fprintf(fout, "%-*s", fieldMax[j], s ? s : "");
+				if (j + 1 < n)
+					fputc('|', fout);
+				if (s)
+					free(s);
+			}
+			fputc('\n', fout);
+		}
+		free(fields);
+		fprintf(fout, "(%d row%s)\n\n", m, m == 1 ? "" : "s");
+		free(fieldMax);
+		free((void *) fieldNames);
+	}
+}
+
 /* checks connection validity */
 static int
 check_cnx_obj(pgobject * self)
 {
 	if (!self->valid)
 	{
-		PyErr_SetString(IntegrityError, "connection has been closed");
+		PyErr_SetString(IntegrityError, "connection has been closed.");
 		return 0;
 	}
 	return 1;
@@ -289,7 +385,7 @@ get_type_array(PGresult *result, int nfields)
 
 	if ((typ = malloc(sizeof(int) * nfields)) == NULL)
 	{
-		PyErr_SetString(PyExc_SyntaxError, "memory error in getresult().");
+		PyErr_SetString(PyExc_MemoryError, "memory error in getresult().");
 		return NULL;
 	}
 
@@ -407,7 +503,6 @@ pgsource_execute(pgsourceobject * self, PyObject * args)
 {
 	char	   *query;
 	const char *temp;
-	long		num_rows;
 
 	/* checks validity */
 	if (!check_source_obj(self, CHECK_CNX))
@@ -449,6 +544,7 @@ pgsource_execute(pgsourceobject * self, PyObject * args)
 	/* checks result status */
 	switch (PQresultStatus(self->last_result))
 	{
+			long		num_rows;
 
 			/* query succeeded */
 		case PGRES_TUPLES_OK:	/* DQL: returns None (DB-SIG compliant) */
@@ -462,7 +558,7 @@ pgsource_execute(pgsourceobject * self, PyObject * args)
 		case PGRES_COPY_IN:
 			self->result_type = RESULT_DDL;
 			temp = PQcmdTuples(self->last_result);
-			num_rows = 1;
+			num_rows = -1;
 			if (temp[0] != 0)
 			{
 				self->result_type = RESULT_DML;
@@ -481,7 +577,7 @@ pgsource_execute(pgsourceobject * self, PyObject * args)
 			break;
 		default:
 			PyErr_SetString(InternalError, "internal error: "
-							"unknown result status.");
+				"unknown result status.");
 			break;
 	}
 
@@ -508,8 +604,8 @@ pgsource_oidstatus(pgsourceobject * self, PyObject * args)
 	/* checks args */
 	if ((args != NULL) && (!PyArg_ParseTuple(args, "")))
 	{
-		PyErr_SetString(PyExc_SyntaxError,
-						"method oidstatus() takes no parameters.");
+		PyErr_SetString(PyExc_TypeError,
+			"method oidstatus() takes no parameters.");
 		return NULL;
 	}
 
@@ -544,8 +640,8 @@ pgsource_fetch(pgsourceobject * self, PyObject * args)
 	size = self->arraysize;
 	if (!PyArg_ParseTuple(args, "|l", &size))
 	{
-		PyErr_SetString(PyExc_SyntaxError,
-						"fetch(num), with num (integer, optional).");
+		PyErr_SetString(PyExc_TypeError,
+			"fetch(num), with num (integer, optional).");
 		return NULL;
 	}
 
@@ -600,15 +696,9 @@ pgsource_move(pgsourceobject * self, PyObject * args, int move)
 	if (!PyArg_ParseTuple(args, ""))
 	{
 		char		errbuf[256];
-
-#ifdef	NO_SNPRINTF
-		sprintf(errbuf, "method %s() takes no parameter.", __movename[move]);
-#else
-		snprintf(errbuf, sizeof(errbuf),
-				 "method %s() takes no parameter.", __movename[move]);
-#endif
-
-		PyErr_SetString(PyExc_SyntaxError, errbuf);
+		PyOS_snprintf(errbuf, sizeof(errbuf),
+			"method %s() takes no parameter.", __movename[move]);
+		PyErr_SetString(PyExc_TypeError, errbuf);
 		return NULL;
 	}
 
@@ -747,7 +837,7 @@ pgsource_listinfo(pgsourceobject * self, PyObject * args)
 	if (!PyArg_ParseTuple(args, ""))
 	{
 		PyErr_SetString(PyExc_TypeError,
-						"method listinfo() takes no parameter.");
+			"method listinfo() takes no parameter.");
 		return NULL;
 	}
 
@@ -933,20 +1023,14 @@ pgsource_setattr(pgsourceobject * self, char *name, PyObject * v)
 }
 
 /* prints query object in human readable format */
+
 static int
 pgsource_print(pgsourceobject * self, FILE *fp, int flags)
 {
-	PQprintOpt	op;
-
 	switch (self->result_type)
 	{
 		case RESULT_DQL:
-			memset(&op, 0, sizeof(op));
-			op.align = 1;
-			op.header = 1;
-			op.fieldSep = "|";
-			op.pager = 1;
-			PQprint(fp, self->last_result, &op);
+			print_result(fp, self->last_result);
 			break;
 		case RESULT_DDL:
 		case RESULT_DML:
@@ -1060,8 +1144,8 @@ pglarge_close(pglargeobject * self, PyObject * args)
 	/* checks args */
 	if (!PyArg_ParseTuple(args, ""))
 	{
-		PyErr_SetString(PyExc_SyntaxError,
-						"method close() takes no parameters.");
+		PyErr_SetString(PyExc_TypeError,
+			"method close() takes no parameters.");
 		return NULL;
 	}
 
@@ -1145,7 +1229,7 @@ pglarge_write(pglargeobject * self, PyObject * args)
 	if (!PyArg_ParseTuple(args, "s#", &buffer, &bufsize))
 	{
 		PyErr_SetString(PyExc_TypeError,
-						"write(buffer), with buffer (sized string).");
+			"write(buffer), with buffer (sized string).");
 		return NULL;
 	}
 
@@ -1213,8 +1297,8 @@ pglarge_size(pglargeobject * self, PyObject * args)
 	/* checks args */
 	if (!PyArg_ParseTuple(args, ""))
 	{
-		PyErr_SetString(PyExc_SyntaxError,
-						"method size() takes no parameters.");
+		PyErr_SetString(PyExc_TypeError,
+			"method size() takes no parameters.");
 		return NULL;
 	}
 
@@ -1240,7 +1324,7 @@ pglarge_size(pglargeobject * self, PyObject * args)
 	if ((start = lo_lseek(self->pgcnx->cnx, self->lo_fd, start, SEEK_SET)) == -1)
 	{
 		PyErr_SetString(PyExc_IOError,
-						"error while moving back to first position.");
+			"error while moving back to first position.");
 		return NULL;
 	}
 
@@ -1261,8 +1345,8 @@ pglarge_tell(pglargeobject * self, PyObject * args)
 	/* checks args */
 	if (!PyArg_ParseTuple(args, ""))
 	{
-		PyErr_SetString(PyExc_SyntaxError,
-						"method tell() takes no parameters.");
+		PyErr_SetString(PyExc_TypeError,
+			"method tell() takes no parameters.");
 		return NULL;
 	}
 
@@ -1299,7 +1383,7 @@ pglarge_export(pglargeobject * self, PyObject * args)
 	if (!PyArg_ParseTuple(args, "s", &name))
 	{
 		PyErr_SetString(PyExc_TypeError,
-						"export(filename), with filename (string).");
+			"export(filename), with filename (string).");
 		return NULL;
 	}
 
@@ -1325,8 +1409,8 @@ pglarge_unlink(pglargeobject * self, PyObject * args)
 	/* checks args */
 	if (!PyArg_ParseTuple(args, ""))
 	{
-		PyErr_SetString(PyExc_SyntaxError,
-						"method unlink() takes no parameters.");
+		PyErr_SetString(PyExc_TypeError,
+			"method unlink() takes no parameters.");
 		return NULL;
 	}
 
@@ -1425,30 +1509,11 @@ static int
 pglarge_print(pglargeobject * self, FILE *fp, int flags)
 {
 	char		print_buffer[128];
-
-	if (self->lo_fd >= 0)
-	{
-#ifdef	NO_SNPRINTF
-		sprintf(print_buffer,
-				"Opened large object, oid %ld", (long) self->lo_oid);
-#else
-		snprintf(print_buffer, sizeof(print_buffer),
-				 "Opened large object, oid %ld", (long) self->lo_oid);
-#endif
-		fputs(print_buffer, fp);
-	}
-	else
-	{
-#ifdef	NO_SNPRINTF
-		sprintf(print_buffer,
-				"Closed large object, oid %ld", (long) self->lo_oid);
-#else
-		snprintf(print_buffer, sizeof(print_buffer),
-				 "Closed large object, oid %ld", (long) self->lo_oid);
-#endif
-		fputs(print_buffer, fp);
-	}
-
+	PyOS_snprintf(print_buffer, sizeof(print_buffer),
+		self->lo_fd >= 0 ?
+			"Opened large object, oid %ld" :
+			"Closed large object, oid %ld", (long) self->lo_oid);
+	fputs(print_buffer, fp);
 	return 0;
 }
 
@@ -1620,15 +1685,15 @@ pg_reset(pgobject * self, PyObject * args)
 {
 	if (!self->cnx)
 	{
-		PyErr_SetString(PyExc_TypeError, "Connection is not valid");
+		PyErr_SetString(PyExc_TypeError, "Connection is not valid.");
 		return NULL;
 	}
 
 	/* checks args */
 	if (!PyArg_ParseTuple(args, ""))
 	{
-		PyErr_SetString(PyExc_SyntaxError,
-						"method reset() takes no parameters.");
+		PyErr_SetString(PyExc_TypeError,
+			"method reset() takes no parameters.");
 		return NULL;
 	}
 
@@ -1638,6 +1703,31 @@ pg_reset(pgobject * self, PyObject * args)
 	return Py_None;
 }
 
+/* cancels current command */
+static char pg_cancel__doc__[] =
+"cancel() -- abandon processing of the current command.";
+
+static PyObject *
+pg_cancel(pgobject * self, PyObject * args)
+{
+	if (!self->cnx)
+	{
+		PyErr_SetString(PyExc_TypeError, "Connection is not valid.");
+		return NULL;
+	}
+
+	/* checks args */
+	if (!PyArg_ParseTuple(args, ""))
+	{
+		PyErr_SetString(PyExc_TypeError,
+			"method reset() takes no parameters.");
+		return NULL;
+	}
+
+	/* request that the server abandon processing of the current command */
+    return PyInt_FromLong((long) PQrequestCancel(self->cnx));
+}
+
 /* get connection socket */
 static char pg_fileno__doc__[] =
 "fileno() -- return database connection socket file handle.";
@@ -1647,15 +1737,15 @@ pg_fileno(pgobject * self, PyObject * args)
 {
 	if (!self->cnx)
 	{
-		PyErr_SetString(PyExc_TypeError, "Connection is not valid");
+		PyErr_SetString(PyExc_TypeError, "Connection is not valid.");
 		return NULL;
 	}
 
 	/* checks args */
 	if (!PyArg_ParseTuple(args, ""))
 	{
-		PyErr_SetString(PyExc_SyntaxError,
-						"method fileno() takes no parameters.");
+		PyErr_SetString(PyExc_TypeError,
+			"method fileno() takes no parameters.");
 		return NULL;
 	}
 
@@ -1676,8 +1766,8 @@ pgquery_ntuples(pgqueryobject * self, PyObject * args)
 	/* checks args */
 	if (!PyArg_ParseTuple(args, ""))
 	{
-		PyErr_SetString(PyExc_SyntaxError,
-						"method ntuples() takes no parameters.");
+		PyErr_SetString(PyExc_TypeError,
+			"method ntuples() takes no parameters.");
 		return NULL;
 	}
 
@@ -1700,8 +1790,8 @@ pgquery_listfields(pgqueryobject * self, PyObject * args)
 	/* checks args */
 	if (!PyArg_ParseTuple(args, ""))
 	{
-		PyErr_SetString(PyExc_SyntaxError,
-						"method listfields() takes no parameters.");
+		PyErr_SetString(PyExc_TypeError,
+			"method listfields() takes no parameters.");
 		return NULL;
 	}
 
@@ -1733,7 +1823,7 @@ pgquery_fieldname(pgqueryobject * self, PyObject * args)
 	if (!PyArg_ParseTuple(args, "i", &i))
 	{
 		PyErr_SetString(PyExc_TypeError,
-						"fieldname(number), with number(integer).");
+			"fieldname(number), with number(integer).");
 		return NULL;
 	}
 
@@ -1797,8 +1887,8 @@ pgquery_getresult(pgqueryobject * self, PyObject * args)
 	/* checks args (args == NULL for an internal call) */
 	if ((args != NULL) && (!PyArg_ParseTuple(args, "")))
 	{
-		PyErr_SetString(PyExc_SyntaxError,
-						"method getresult() takes no parameters.");
+		PyErr_SetString(PyExc_TypeError,
+			"method getresult() takes no parameters.");
 		return NULL;
 	}
 
@@ -1823,6 +1913,7 @@ pgquery_getresult(pgqueryobject * self, PyObject * args)
 			int			k;
 			char	   *s = PQgetvalue(self->last_result, i, j);
 			char		cashbuf[64];
+			PyObject   *float_str;
 
 			if (PQgetisnull(self->last_result, i, j))
 			{
@@ -1841,7 +1932,9 @@ pgquery_getresult(pgqueryobject * self, PyObject * args)
 						break;
 
 					case 3:
-						val = PyFloat_FromString(PyString_FromString(s), NULL);
+						float_str = PyString_FromString(s);
+						val = PyFloat_FromString(float_str, NULL);
+						Py_DECREF(float_str);
 						break;
 
 					case 4:
@@ -1919,8 +2012,8 @@ pgquery_dictresult(pgqueryobject * self, PyObject * args)
 	/* checks args (args == NULL for an internal call) */
 	if ((args != NULL) && (!PyArg_ParseTuple(args, "")))
 	{
-		PyErr_SetString(PyExc_SyntaxError,
-						"method getresult() takes no parameters.");
+		PyErr_SetString(PyExc_TypeError,
+			"method getresult() takes no parameters.");
 		return NULL;
 	}
 
@@ -1945,6 +2038,7 @@ pgquery_dictresult(pgqueryobject * self, PyObject * args)
 			int			k;
 			char	   *s = PQgetvalue(self->last_result, i, j);
 			char		cashbuf[64];
+			PyObject   *float_str;
 
 			if (PQgetisnull(self->last_result, i, j))
 			{
@@ -1963,7 +2057,9 @@ pgquery_dictresult(pgqueryobject * self, PyObject * args)
 						break;
 
 					case 3:
-						val = PyFloat_FromString(PyString_FromString(s), NULL);
+						float_str = PyString_FromString(s);
+						val = PyFloat_FromString(float_str, NULL);
+						Py_DECREF(float_str);
 						break;
 
 					case 4:
@@ -2039,15 +2135,15 @@ pg_getnotify(pgobject * self, PyObject * args)
 
 	if (!self->cnx)
 	{
-		PyErr_SetString(PyExc_TypeError, "Connection is not valid");
+		PyErr_SetString(PyExc_TypeError, "Connection is not valid.");
 		return NULL;
 	}
 
 	/* checks args */
 	if (!PyArg_ParseTuple(args, ""))
 	{
-		PyErr_SetString(PyExc_SyntaxError,
-						"method getnotify() takes no parameters.");
+		PyErr_SetString(PyExc_TypeError,
+			"method getnotify() takes no parameters.");
 		return NULL;
 	}
 
@@ -2136,7 +2232,7 @@ pg_query(pgobject * self, PyObject * args)
 
 	if (!self->cnx)
 	{
-		PyErr_SetString(PyExc_TypeError, "Connection is not valid");
+		PyErr_SetString(PyExc_TypeError, "Connection is not valid.");
 		return NULL;
 	}
 
@@ -2199,7 +2295,7 @@ pg_query(pgobject * self, PyObject * args)
 				return Py_None;
 			default:
 				PyErr_SetString(InternalError, "internal error: "
-								"unknown result status.");
+					"unknown result status.");
 				break;
 		}
 
@@ -2226,7 +2322,7 @@ pg_putline(pgobject * self, PyObject * args)
 
 	if (!self->cnx)
 	{
-		PyErr_SetString(PyExc_TypeError, "Connection is not valid");
+		PyErr_SetString(PyExc_TypeError, "Connection is not valid.");
 		return NULL;
 	}
 
@@ -2255,15 +2351,15 @@ pg_getline(pgobject * self, PyObject * args)
 
 	if (!self->cnx)
 	{
-		PyErr_SetString(PyExc_TypeError, "Connection is not valid");
+		PyErr_SetString(PyExc_TypeError, "Connection is not valid.");
 		return NULL;
 	}
 
 	/* checks args */
 	if (!PyArg_ParseTuple(args, ""))
 	{
-		PyErr_SetString(PyExc_SyntaxError,
-						"method getline() takes no parameters.");
+		PyErr_SetString(PyExc_TypeError,
+			"method getline() takes no parameters.");
 		return NULL;
 	}
 
@@ -2295,15 +2391,15 @@ pg_endcopy(pgobject * self, PyObject * args)
 {
 	if (!self->cnx)
 	{
-		PyErr_SetString(PyExc_TypeError, "Connection is not valid");
+		PyErr_SetString(PyExc_TypeError, "Connection is not valid.");
 		return NULL;
 	}
 
 	/* checks args */
 	if (!PyArg_ParseTuple(args, ""))
 	{
-		PyErr_SetString(PyExc_SyntaxError,
-						"method endcopy() takes no parameters.");
+		PyErr_SetString(PyExc_TypeError,
+			"method endcopy() takes no parameters.");
 		return NULL;
 	}
 
@@ -2318,14 +2414,7 @@ pg_endcopy(pgobject * self, PyObject * args)
 static PyObject *
 pgquery_print(pgqueryobject * self, FILE *fp, int flags)
 {
-	PQprintOpt	op;
-
-	memset(&op, 0, sizeof(op));
-	op.align = 1;
-	op.header = 1;
-	op.fieldSep = "|";
-	op.pager = 1;
-	PQprint(fp, self->last_result, &op);
+	print_result(fp, self->last_result);
 	return 0;
 }
 
@@ -2346,8 +2435,7 @@ pg_inserttable(pgobject * self, PyObject * args)
 	PGresult	*result;
 	char		*table,
 				*buffer,
-				*bufpt,
-				*temp;
+				*bufpt;
 	size_t		bufsiz;
 	PyObject	*list,
 				*sublist,
@@ -2361,7 +2449,7 @@ pg_inserttable(pgobject * self, PyObject * args)
 
 	if (!self->cnx)
 	{
-		PyErr_SetString(PyExc_TypeError, "Connection is not valid");
+		PyErr_SetString(PyExc_TypeError, "Connection is not valid.");
 		return NULL;
 	}
 
@@ -2369,7 +2457,7 @@ pg_inserttable(pgobject * self, PyObject * args)
 	if (!PyArg_ParseTuple(args, "sO:filter", &table, &list))
 	{
 		PyErr_SetString(PyExc_TypeError,
-			"tableinsert(table, content), with table (string) "
+			"inserttable(table, content), with table (string) "
 			"and content (list).");
 		return NULL;
 	}
@@ -2392,181 +2480,150 @@ pg_inserttable(pgobject * self, PyObject * args)
 		return NULL;
 	}
 
-	if (m)
-	{
-		/* not strictly necessary but removes a bogus warning */
-		n = 0;
+    /* allocate buffer */
+    if (!(buffer = malloc(MAX_BUFFER_SIZE)))
+    {
+        PyErr_SetString(PyExc_MemoryError,
+            "can't allocate insert buffer.");
+        return NULL;
+    }
 
-		/* checks sublists type and size */
-		for (i = 0; i < m; i++)
-		{
-			sublist = getitem(list, i);
-			if (PyTuple_Check(sublist))
-				j = PyTuple_Size(sublist);
-			else if (PyList_Check(sublist))
-				j = PyList_Size(sublist);
-			else
-			{
-				PyErr_SetString(PyExc_TypeError,
-					"second arg must contain some kind of arrays.");
-				return NULL;
-			}
-			if (i)
-			{
-				if (j != n)
-				{
-					PyErr_SetString(PyExc_TypeError,
-						"arrays contained in second arg must have same size.");
-					return NULL;
-				}
-			}
-			else
-				n = j;		/* never used before this assignment */
-		}
-		if (n)
-		{	
-			/* allocate buffer */
-			if (!(buffer = malloc(MAX_BUFFER_SIZE)))
-			{		
-				PyErr_SetString(PyExc_MemoryError,
-					"can't allocate insert buffer.");
-				return NULL;
-			}
+    /* starts query */
+    sprintf(buffer, "copy %s from stdin", table);
 
-			/* starts query */
-			sprintf(buffer, "copy %s from stdin", table);
+    Py_BEGIN_ALLOW_THREADS
+        result = PQexec(self->cnx, buffer);
+    Py_END_ALLOW_THREADS
 
-			Py_BEGIN_ALLOW_THREADS
-				result = PQexec(self->cnx, buffer);
-			Py_END_ALLOW_THREADS
+    if (!result)
+    {
+        free(buffer);
+        PyErr_SetString(PyExc_ValueError, PQerrorMessage(self->cnx));
+        return NULL;
+    }
 
-			if (!result)
-			{
-				free(buffer);
-				PyErr_SetString(PyExc_ValueError, PQerrorMessage(self->cnx));
-				return NULL;
-			}
+    PQclear(result);
 
-			PQclear(result);
+    n = 0; /* not strictly necessary but avoids warning */
 
-			/* feeds table */
-			for (i = 0; i < m; i++)
-			{
-				sublist = getitem(list, i);
-				if (PyTuple_Check(sublist))
-					getsubitem = PyTuple_GetItem;
-				else
-					getsubitem = PyList_GetItem;
-	
-				/* builds insert line */
-				bufpt=buffer;
-				bufsiz = MAX_BUFFER_SIZE - 1;
+    /* feed table */
+    for (i = 0; i < m; i++)
+    {
+        sublist = getitem(list, i);
+        if (PyTuple_Check(sublist))
+        {
+            j = PyTuple_Size(sublist);
+            getsubitem = PyTuple_GetItem;
+        }
+        else if (PyList_Check(sublist))
+        {
+            j = PyList_Size(sublist);
+            getsubitem = PyList_GetItem;
+        }
+        else
+        {
+            PyErr_SetString(PyExc_TypeError,
+                "second arg must contain some kind of arrays.");
+            return NULL;
+        }
+        if (i)
+        {
+            if (j != n)
+            {
+                free(buffer);
+                PyErr_SetString(PyExc_TypeError,
+                    "arrays contained in second arg must have same size.");
+                return NULL;
+            }
+        }
+        else
+        {
+            n = j; /* never used before this assignment */
+        }
 
-				for (j = 0; j < n; j++)
-				{
-					item = getsubitem(sublist, j);
-					/* converts item to string */
-					if (item == Py_None)
-					{
-						if (bufsiz > 2)
-						{
-							*bufpt++ = '\\'; *bufpt++ = 'N';
-							bufsiz -= 2;
-						}
-						else
-							bufsiz = 0;
-					}
-					else if (PyString_Check(item))
-					{
-						temp = PyString_AS_STRING(item);
-						while (*temp && bufsiz)
-						{
-							if (*temp == '\\' || *temp == '\t' || *temp == '\n')
-							{
-								*bufpt++='\\'; --bufsiz;
-								if (!bufsiz) break;
-							}
-							*bufpt++=*temp++; --bufsiz;
-						}
-					}
-					else if (PyInt_Check(item))
-					{
-						long	k;
-						int 	h;
+        /* builds insert line */
+        bufpt = buffer;
+        bufsiz = MAX_BUFFER_SIZE - 1;
 
-						k = PyInt_AsLong(item);
-						h = snprintf(bufpt, bufsiz, "%ld", k);
-						if (h > 0)
-						{
-							bufpt += h; bufsiz -= h;
-						}
-						else
-							bufsiz = 0;
-					}
-					else if (PyLong_Check(item))
-					{
-						long	k;
-						int 	h;
+        for (j = 0; j < n; j++)
+        {
+            if (j)
+            {
+                *bufpt++ = '\t'; --bufsiz;
+            }
 
-						k = PyLong_AsLong(item);
-						h = snprintf(bufpt, bufsiz, "%ld", k);
-						if (h > 0)
-						{
-							bufpt += h; bufsiz -= h;
-						}
-						else
-							bufsiz = 0;
-					}
-					else if (PyFloat_Check(item))
-					{
-						double	k;
-						int		h;
+            item = getsubitem(sublist, j);
 
-						k = PyFloat_AS_DOUBLE(item);
-						h = snprintf(bufpt, bufsiz, "%g", k);
-						if (h > 0)
-						{
-							bufpt += h; bufsiz -= h;
-						}
-						else
-							bufsiz = 0;
-					}
-					else
-					{
-						free(buffer);
-						PyErr_SetString(PyExc_ValueError,
-							"items must be strings, integers, "
-							"longs or double (real).");
-						return NULL;
-					}
+            /* convert item to string and append to buffer */
+            if (item == Py_None)
+            {
+                if (bufsiz > 2)
+                {
+                    *bufpt++ = '\\'; *bufpt++ = 'N';
+                    bufsiz -= 2;
+                }
+                else
+                    bufsiz = 0;
+            }
+            else if (PyString_Check(item))
+            {
+                const char* t = PyString_AS_STRING(item);
+                while (*t && bufsiz)
+                {
+                    if (*t == '\\' || *t == '\t' || *t == '\n')
+                    {
+                        *bufpt++ = '\\'; --bufsiz;
+                        if (!bufsiz) break;
+                    }
+                    *bufpt++ = *t++; --bufsiz;
+                }
+            }
+            else if (PyInt_Check(item) || PyLong_Check(item))
+            {
+                PyObject* s = PyObject_Str(item);
+                const char* t = PyString_AsString(s);
+                while (*t && bufsiz)
+                {
+                    *bufpt++ = *t++; --bufsiz;
+                }
+                Py_DECREF(s);
+            }
+            else
+            {
+                PyObject* s = PyObject_Repr(item);
+                const char* t = PyString_AsString(s);
+                while (*t && bufsiz)
+                {
+                    if (*t == '\\' || *t == '\t' || *t == '\n')
+                    {
+                        *bufpt++ = '\\'; --bufsiz;
+                        if (!bufsiz) break;
+                    }
+                    *bufpt++ = *t++; --bufsiz;
+                }
+                Py_DECREF(s);
+            }
 
-					if (j < n-1)
-					{
-						*bufpt++ = '\t'; --bufsiz;
-					}
-			
-					if (bufsiz <= 0)
-					{
-						free(buffer);
-						PyErr_SetString(PyExc_MemoryError,
-							"insert buffer overflow.");
-						return NULL;
-					}
+            if (bufsiz <= 0)
+            {
+                free(buffer);
+                PyErr_SetString(PyExc_MemoryError,
+                    "insert buffer overflow.");
+                return NULL;
+            }
 
-				}
+        }
 
-				*bufpt++ = '\n'; *bufpt = '\0';
+        *bufpt++ = '\n'; *bufpt = '\0';
 
-				/* sends data */
-				PQputline(self->cnx, buffer);
-			}
-	
-			/* ends query */
-			PQputline(self->cnx, "\\.\n");
-			PQendcopy(self->cnx);
-			free(buffer);
-		}
-	}
+        /* sends data */
+        PQputline(self->cnx, buffer);
+    }
+
+    /* ends query */
+    PQputline(self->cnx, "\\.\n");
+    PQendcopy(self->cnx);
+    free(buffer);
 
 	/* no error : returns nothing */
 	Py_INCREF(Py_None);
@@ -2583,15 +2640,15 @@ pg_transaction(pgobject * self, PyObject * args)
 {
 	if (!self->cnx)
 	{
-		PyErr_SetString(PyExc_TypeError, "Connection is not valid");
+		PyErr_SetString(PyExc_TypeError, "Connection is not valid.");
 		return NULL;
 	}
 
 	/* checks args */
 	if (!PyArg_ParseTuple(args, ""))
 	{
-		PyErr_SetString(PyExc_SyntaxError,
-						"method transaction() takes no parameters.");
+		PyErr_SetString(PyExc_TypeError,
+			"method transaction() takes no parameters.");
 		return NULL;
 	}
 
@@ -2618,7 +2675,7 @@ pg_locreate(pgobject * self, PyObject * args)
 	if (!PyArg_ParseTuple(args, "i", &mode))
 	{
 		PyErr_SetString(PyExc_TypeError,
-						"locreate(mode), with mode (integer).");
+			"locreate(mode), with mode (integer).");
 		return NULL;
 	}
 
@@ -2702,6 +2759,7 @@ static struct PyMethodDef pgobj_methods[] = {
 	{"source", (PyCFunction) pg_source, METH_VARARGS, pg_source__doc__},
 	{"query", (PyCFunction) pg_query, METH_VARARGS, pg_query__doc__},
 	{"reset", (PyCFunction) pg_reset, METH_VARARGS, pg_reset__doc__},
+    {"cancel", (PyCFunction) pg_cancel, METH_VARARGS, pg_cancel__doc__},
 	{"close", (PyCFunction) pg_close, METH_VARARGS, pg_close__doc__},
 	{"fileno", (PyCFunction) pg_fileno, METH_VARARGS, pg_fileno__doc__},
 	{"getnotify", (PyCFunction) pg_getnotify, METH_VARARGS,
@@ -2741,7 +2799,7 @@ pg_getattr(pgobject * self, char *name)
 	/* first exception - close which returns a different error */
 	if (strcmp(name, "close") && !self->cnx)
 	{
-		PyErr_SetString(PyExc_TypeError, "Connection is not valid");
+		PyErr_SetString(PyExc_TypeError, "Connection is not valid.");
 		return NULL;
 	}
 
@@ -2879,6 +2937,80 @@ staticforward PyTypeObject PgQueryType = {
 
 /* MODULE FUNCTIONS */
 
+/* escape string */
+static char escape_string__doc__[] =
+"escape_string(str) -- escape a string for use within SQL.";
+
+static PyObject *
+escape_string(PyObject *self, PyObject *args) {
+	char *from; /* our string argument */
+	char *to=NULL; /* the result */
+	int from_length; /* length of string */
+	int to_length; /* length of result */
+	PyObject *ret; /* string object to return */
+
+	if (!PyArg_ParseTuple(args, "s#", &from, &from_length))
+		return NULL;
+    to_length = 2*from_length + 1;
+    if (to_length < from_length) { /* overflow */
+        to_length = from_length;
+        from_length = (from_length - 1)/2;
+    }
+	to = (char *)malloc(to_length);
+	to_length = (int)PQescapeString(to, from, (size_t)from_length);
+	ret = Py_BuildValue("s#", to, to_length);
+    if (to)
+        free(to);
+	if (!ret) /* pass on exception */
+        return NULL;
+	return ret;
+}
+
+/* escape bytea */
+static char escape_bytea__doc__[] =
+"escape_bytea(data) -- escape binary data for use within SQL as type bytea.";
+
+static PyObject *
+escape_bytea(PyObject *self, PyObject *args) {
+	unsigned char *from; /* our string argument */
+	unsigned char *to; /* the result */
+	int from_length; /* length of string */
+	size_t to_length; /* length of result */
+	PyObject *ret; /* string object to return */
+
+	if (!PyArg_ParseTuple(args, "s#", &from, &from_length))
+		return NULL;
+	to = PQescapeBytea(from, (int)from_length, &to_length);
+	ret = Py_BuildValue("s", to);
+    if (to)
+        PQfreemem((void *)to);
+	if (!ret) /* pass on exception */
+        return NULL;
+	return ret;
+}
+
+/* unescape bytea */
+static char unescape_bytea__doc__[] =
+"unescape_bytea(str) -- unescape bytea data that has been retrieved as text.";
+
+static PyObject
+*unescape_bytea(PyObject *self, PyObject *args) {
+	unsigned char *from; /* our string argument */
+	unsigned char *to; /* the result */
+	size_t to_length; /* length of result string */
+	PyObject *ret; /* string object to return */
+
+	if (!PyArg_ParseTuple(args, "s", &from))
+		return NULL;
+	to = PQunescapeBytea(from, &to_length);
+	ret = Py_BuildValue("s#", to, (int)to_length);
+    if (to)
+        PQfreemem((void *)to);
+	if (!ret) /* pass on exception */
+        return NULL;
+	return ret;
+}
+
 #ifdef DEFAULT_VARS
 
 /* gets default host */
@@ -2891,8 +3023,8 @@ pggetdefhost(PyObject * self, PyObject * args)
 	/* checks args */
 	if (!PyArg_ParseTuple(args, ""))
 	{
-		PyErr_SetString(PyExc_SyntaxError,
-						"method get_defhost() takes no parameter.");
+		PyErr_SetString(PyExc_TypeError,
+			"method get_defhost() takes no parameter.");
 		return NULL;
 	}
 
@@ -2914,7 +3046,7 @@ pgsetdefhost(PyObject * self, PyObject * args)
 	if (!PyArg_ParseTuple(args, "z", &temp))
 	{
 		PyErr_SetString(PyExc_TypeError,
-						"set_defhost(name), with name (string/None).");
+			"set_defhost(name), with name (string/None).");
 		return NULL;
 	}
 
@@ -2942,8 +3074,8 @@ pggetdefbase(PyObject * self, PyObject * args)
 	/* checks args */
 	if (!PyArg_ParseTuple(args, ""))
 	{
-		PyErr_SetString(PyExc_SyntaxError,
-						"method get_defbase() takes no parameter.");
+		PyErr_SetString(PyExc_TypeError,
+			"method get_defbase() takes no parameter.");
 		return NULL;
 	}
 
@@ -2965,7 +3097,7 @@ pgsetdefbase(PyObject * self, PyObject * args)
 	if (!PyArg_ParseTuple(args, "z", &temp))
 	{
 		PyErr_SetString(PyExc_TypeError,
-						"set_defbase(name), with name (string/None).");
+			"set_defbase(name), with name (string/None).");
 		return NULL;
 	}
 
@@ -2993,8 +3125,8 @@ pggetdefopt(PyObject * self, PyObject * args)
 	/* checks args */
 	if (!PyArg_ParseTuple(args, ""))
 	{
-		PyErr_SetString(PyExc_SyntaxError,
-						"method get_defopt() takes no parameter.");
+		PyErr_SetString(PyExc_TypeError,
+			"method get_defopt() takes no parameter.");
 		return NULL;
 	}
 
@@ -3016,7 +3148,7 @@ pgsetdefopt(PyObject * self, PyObject * args)
 	if (!PyArg_ParseTuple(args, "z", &temp))
 	{
 		PyErr_SetString(PyExc_TypeError,
-						"set_defopt(name), with name (string/None).");
+			"set_defopt(name), with name (string/None).");
 		return NULL;
 	}
 
@@ -3044,8 +3176,8 @@ pggetdeftty(PyObject * self, PyObject * args)
 	/* checks args */
 	if (!PyArg_ParseTuple(args, ""))
 	{
-		PyErr_SetString(PyExc_SyntaxError,
-						"method get_deftty() takes no parameter.");
+		PyErr_SetString(PyExc_TypeError,
+			"method get_deftty() takes no parameter.");
 		return NULL;
 	}
 
@@ -3068,7 +3200,7 @@ pgsetdeftty(PyObject * self, PyObject * args)
 	if (!PyArg_ParseTuple(args, "z", &temp))
 	{
 		PyErr_SetString(PyExc_TypeError,
-						"set_deftty(name), with name (string/None).");
+			"set_deftty(name), with name (string/None).");
 		return NULL;
 	}
 
@@ -3096,8 +3228,8 @@ pggetdefuser(PyObject * self, PyObject * args)
 	/* checks args */
 	if (!PyArg_ParseTuple(args, ""))
 	{
-		PyErr_SetString(PyExc_SyntaxError,
-						"method get_defuser() takes no parameter.");
+		PyErr_SetString(PyExc_TypeError,
+			"method get_defuser() takes no parameter.");
 
 		return NULL;
 	}
@@ -3120,7 +3252,7 @@ pgsetdefuser(PyObject * self, PyObject * args)
 	if (!PyArg_ParseTuple(args, "z", &temp))
 	{
 		PyErr_SetString(PyExc_TypeError,
-						"set_defuser(name), with name (string/None).");
+			"set_defuser(name), with name (string/None).");
 		return NULL;
 	}
 
@@ -3152,7 +3284,7 @@ pgsetdefpasswd(PyObject * self, PyObject * args)
 	if (!PyArg_ParseTuple(args, "z", &temp))
 	{
 		PyErr_SetString(PyExc_TypeError,
-				"set_defpasswd(password), with password (string/None).");
+			"set_defpasswd(password), with password (string/None).");
 		return NULL;
 	}
 
@@ -3181,8 +3313,8 @@ pggetdefport(PyObject * self, PyObject * args)
 	/* checks args */
 	if (!PyArg_ParseTuple(args, ""))
 	{
-		PyErr_SetString(PyExc_SyntaxError,
-						"method get_defport() takes no parameter.");
+		PyErr_SetString(PyExc_TypeError,
+			"method get_defport() takes no parameter.");
 		return NULL;
 	}
 
@@ -3204,7 +3336,7 @@ pgsetdefport(PyObject * self, PyObject * args)
 	if ((!PyArg_ParseTuple(args, "l", &port)) || (port < -1))
 	{
 		PyErr_SetString(PyExc_TypeError, "set_defport(port), with port "
-						"(positive integer/-1).");
+			"(positive integer/-1).");
 		return NULL;
 	}
 
@@ -3228,6 +3360,12 @@ pgsetdefport(PyObject * self, PyObject * args)
 static struct PyMethodDef pg_methods[] = {
 	{"connect", (PyCFunction) pgconnect, METH_VARARGS|METH_KEYWORDS,
 			connect__doc__},
+	{"escape_string", (PyCFunction) escape_string, METH_VARARGS,
+			escape_string__doc__},
+	{"escape_bytea", (PyCFunction) escape_bytea, METH_VARARGS,
+			escape_bytea__doc__},
+	{"unescape_bytea", (PyCFunction) unescape_bytea, METH_VARARGS,
+			unescape_bytea__doc__},
 
 #ifdef DEFAULT_VARS
 	{"get_defhost", pggetdefhost, METH_VARARGS, getdefhost__doc__},
